from typing import Optional, Union
import os

# value interceptors are for permanent or contextual feature value overrides, but mostly they are intended for
# permanent ones


class InterceptorValue:
    _val: Optional[object]

    def __init__(self, val: object):
        self._val = val

    def cast(self, expected_type: Optional[str]) -> Union[None, bool, str, float]:
        if expected_type is None or self._val is None:
            return self._val

        if expected_type == 'BOOLEAN':
            return str(self._val).lower().strip() == 'true'
        elif expected_type == 'NUMBER':
            return float(str(self._val))

        return str(self._val)


class ValueInterceptor:
    def intercepted_value(self, feature_key: str) -> Optional[InterceptorValue]:
        pass


class EnvironmentInterceptor(ValueInterceptor):
    _enabled: bool
    # this is an example of an interceptor that when it starts checks the environment for a feature value override
    def __init__(self):
        self._enabled = os.environ.get("FEATUREHUB_POLL_INTERVAL", "false") == "true"

    def sanitize_feature_name(self, feature_key: str) -> str:
        return feature_key.replace(" ", "_")

    def intercepted_value(self, feature_key: str) -> Optional[InterceptorValue]:
        if self._enabled:
            found = os.environ.get(f"FEATUREHUB_{self.sanitize_feature_name(feature_key)}")
            if found is not None:
                return InterceptorValue(found)

        return None




from typing import Optional

import urllib3
import threading
import logging
import sys
import json
import asyncio
from featurehub_sdk.edge_service import EdgeService
from featurehub_sdk.featurehub_repository import FeatureHubRepository
from featurehub_sdk.version import sdk_version

log = logging.getLogger(sys.modules[__name__].__name__)


class PollingEdgeService(EdgeService):
    _interval: int
    _repository: FeatureHubRepository
    _cancel: bool
    _thread: Optional[threading.Timer]
    _client_eval: bool
    _http: urllib3.PoolManager

    def __init__(self, edge_url: str, api_keys: list[str],
                 repository: FeatureHubRepository,
                 interval: int):
        self._interval = interval
        self._repository = repository
        self._cancel = False
        self._thread = None
        self._client_eval = '*' in api_keys[0]
        self._context = None
        self._http = urllib3.PoolManager()

        self._url = f"{edge_url}/features?" + "&".join(map(lambda i: 'apiKey=' + i, api_keys))
        log.debug(f"polling url {self._url}")

    # allow us to update the interval of the current polling edge service
    def update_interval(self, interval: int):
        self._interval = interval
        old_cancel = self._cancel
        self._cancel = False
        if not old_cancel:
            self.poll_with_interval()

    # this does the business, calls the remote service and gets the features back
    async def _get_updates(self):
        # TODO: set timeout of tcp requests to 12 seconds, or give users control over it using environ vars
        url = self._url if self._context is None else f"{self._url}&{self._context}"
        log.log(5, "polling ", url)
        resp = self._http.request(method='GET', url=url, headers={'X-SDK': 'Python', 'X-SDK-Version': sdk_version})
        log.log(5, "polling status", resp.status)
        x = resp.status
        print(f"polling status {resp}")
        if resp.status == 200:
            self._process_successful_results(json.loads(resp.data.decode('utf-8')))
        elif resp.status == 404: # no such key
            self._repository.notify("failed", None)
            self._cancel = True
            log.error("Specified API Key does not exist %s", self._url)
        elif resp.status == 503:
            # dacha is busy, just wait
            return
        # otherwise its likely a transient failure, so keep trying

    # this is essentially a repeating task because it "calls itself"
    # another way to do this is with a separate class that is itself a thread descendant
    # which waits for the requisite time,  then triggers a callback and then essentially does the same thing
    # if we need a repeating task elsewhere, we should consider refactoring this
    async def poll_with_interval(self):
        if not self._cancel:
            await self._get_updates()
            if not self._cancel and self._interval > 0:
                self._thread = threading.Timer(self._interval, self.poll_again)
                self._thread.daemon = True # allow it to just disappear off if the app closes down
                self._thread.start()

    # this ends up being synchronous
    def poll_again(self):
        asyncio.run(self.poll_with_interval())

    # async polls, you can choose not to wait for updates
    # if the interval is zero, this will just issue a get updates and stop
    async def poll(self):
        self._cancel = False
        await self.poll_with_interval()

    def client_evaluated(self):
        return self._client_eval

    def close(self):
        self._cancel = True
        if self._thread is not None:
            self._thread.cancel()
            self._thread = None

    async def context_change(self, header: str):
        old_context = self._context
        self._context = header
        if old_context != header:
            await self._get_updates()

    @property
    def cancelled(self):
        return self._cancel

    # we get returned a bunch of environments for a GET/Poll API so we need to cycle through them
    # the result is different for streaming
    def _process_successful_results(self, data):
        log.log(5, "featurehub polling data was %s", data)
        for feature_apikey in data:
            if feature_apikey:
                self._repository.notify("features", feature_apikey['features'])



import datetime
from typing import Optional

import re

from semver import cmp
import math
from murmurhash2 import murmurhash3
from ipaddress import ip_address, ip_network

from featurehub_sdk.client_context import ClientContext, RolloutStrategyAttributeConditional, \
    RolloutStrategyFieldType, RolloutStrategy, RolloutStrategyAttribute, Applied


class PercentageCalculator:
    def determine_client_percentage(self, percentage_text: str, feature_id: str) -> float:
        pass


class Murmur3PercentageCalculator:
    MAX_PERCENTAGE = 1000000
    SEED = 0

    def determine_client_percentage(self, percentage_text: str, feature_id: str) -> float:
        result = murmurhash3(bytes(percentage_text + feature_id, 'utf-8'), Murmur3PercentageCalculator.SEED)
        return math.floor(result / math.pow(2, 32) * Murmur3PercentageCalculator.MAX_PERCENTAGE)


class StrategyMatcher:
    def match(self, supplied_value: str, attr: RolloutStrategyAttribute) -> bool:
        pass


class MatcherRepository:
    def find_matcher(self, attr: RolloutStrategyAttribute) -> StrategyMatcher:
        pass


class FallthroughMatcher(StrategyMatcher):
    def match(self, supplied_value: str, attr: RolloutStrategyAttribute) -> bool:
        return False


class BooleanMatcher(StrategyMatcher):
    def match(self, supplied_value: str, attr: RolloutStrategyAttribute) -> bool:
        val = 'true' == supplied_value.lower()

        if attr.conditional == RolloutStrategyAttributeConditional.Equals:
            return val == (str(attr.values[0]).lower() == 'true')

        if attr.conditional == RolloutStrategyAttributeConditional.NotEquals:
            return val != (str(attr.values[0]).lower() == 'true')

        return False


class StringMatcher(StrategyMatcher):
    def match(self, supplied_value: str, attr: RolloutStrategyAttribute) -> bool:
        vals = attr.str_values

        # match was only introduced in python 3.10 so...
        if attr.conditional == RolloutStrategyAttributeConditional.Equals:
            return next(filter(lambda x: supplied_value == x, vals), None) is not None
        elif attr.conditional == RolloutStrategyAttributeConditional.EndsWith:
            return next(filter(lambda x: supplied_value.endswith(x), vals), None) is not None
        elif attr.conditional == RolloutStrategyAttributeConditional.StartsWith:
            return next(filter(lambda x: supplied_value.startswith(x), vals), None) is not None
        elif attr.conditional == RolloutStrategyAttributeConditional.Greater:
            return next(filter(lambda x: supplied_value > x, vals), None) is not None
        elif attr.conditional == RolloutStrategyAttributeConditional.GreaterEquals:
            return next(filter(lambda x: supplied_value >= x, vals), None) is not None
        elif attr.conditional == RolloutStrategyAttributeConditional.Less:
            return next(filter(lambda x: supplied_value < x, vals), None) is not None
        elif attr.conditional == RolloutStrategyAttributeConditional.LessEquals:
            return next(filter(lambda x: supplied_value <= x, vals), None) is not None
        elif attr.conditional == RolloutStrategyAttributeConditional.Includes:
            return next(filter(lambda x: x in supplied_value, vals), None) is not None
        elif attr.conditional == RolloutStrategyAttributeConditional.Excludes:
            return next(filter(lambda x: x in supplied_value, vals), None) is None
        elif attr.conditional == RolloutStrategyAttributeConditional.Regex:
            return next(filter(lambda x: re.search(x, supplied_value), vals), None) is not None
        elif attr.conditional == RolloutStrategyAttributeConditional.NotEquals:
            return next(filter(lambda x: supplied_value == x, vals), None) is None

        return False

# we are not doing Date or DateTime matcher but treating those as plain strings


class NumberMatcher(StrategyMatcher):
    def match(self, supplied_value: str, attr: RolloutStrategyAttribute) -> bool:
        try:
            parsed_val = float(supplied_value)

            # these we treat as strings
            if attr.conditional == RolloutStrategyAttributeConditional.EndsWith:
                return next(filter(lambda x: supplied_value.endswith(x), attr.str_values), None) is not None
            elif attr.conditional == RolloutStrategyAttributeConditional.StartsWith:
                return next(filter(lambda x: supplied_value.startswith(x), attr.str_values), None) is not None
            elif attr.conditional == RolloutStrategyAttributeConditional.Regex:
                return next(filter(lambda x: re.search(x, supplied_value), attr.str_values), None) is not None
            else: # the rest we treat as floats
                vals = attr.float_values

                # match was only introduced in python 3.10 so...
                if attr.conditional == RolloutStrategyAttributeConditional.Equals or \
                        attr.conditional == RolloutStrategyAttributeConditional.Includes:
                    return next(filter(lambda x: parsed_val == x, vals), None) is not None
                elif attr.conditional == RolloutStrategyAttributeConditional.Greater:
                    return next(filter(lambda x: parsed_val > x, vals), None) is not None
                elif attr.conditional == RolloutStrategyAttributeConditional.GreaterEquals:
                    return next(filter(lambda x: parsed_val >= x, vals), None) is not None
                elif attr.conditional == RolloutStrategyAttributeConditional.Less:
                    return next(filter(lambda x: parsed_val < x, vals), None) is not None
                elif attr.conditional == RolloutStrategyAttributeConditional.LessEquals:
                    return next(filter(lambda x: parsed_val <= x, vals), None) is not None
                elif attr.conditional == RolloutStrategyAttributeConditional.NotEquals or \
                        attr.conditional == RolloutStrategyAttributeConditional.Excludes:
                    return next(filter(lambda x: parsed_val == x, vals), None) is None
        except ValueError:
            pass

        return False


class SemanticVersionMatcher(StrategyMatcher):
    def match(self, supplied_value: str, attr: RolloutStrategyAttribute) -> bool:
        vals = attr.str_values

        if attr.conditional == RolloutStrategyAttributeConditional.Includes \
            or attr.conditional == RolloutStrategyAttributeConditional.Equals:
            return next(filter(lambda x: cmp(supplied_value, "==", x, True), vals), None) is not None
        elif attr.conditional == RolloutStrategyAttributeConditional.Excludes \
            or attr.conditional == RolloutStrategyAttributeConditional.NotEquals:
            return next(filter(lambda x: cmp(supplied_value, "==", x, True), vals), None) is None
        elif attr.conditional == RolloutStrategyAttributeConditional.Greater:
            return next(filter(lambda x: cmp(supplied_value, ">", x, True), vals), None) is not None
        elif attr.conditional == RolloutStrategyAttributeConditional.GreaterEquals:
            return next(filter(lambda x: cmp(supplied_value, ">=", x, True), vals), None) is not None
        elif attr.conditional == RolloutStrategyAttributeConditional.Less:
            return next(filter(lambda x: cmp(supplied_value, "<", x, True), vals), None) is not None
        elif attr.conditional == RolloutStrategyAttributeConditional.LessEquals:
            return next(filter(lambda x: cmp(supplied_value, "<=", x, True), vals), None) is not None

        return False


class IPNetworkMatcher(StrategyMatcher):
    def match(self, supplied_value: str, attr: RolloutStrategyAttribute) -> bool:

        if attr.conditional == RolloutStrategyAttributeConditional.Includes \
                or attr.conditional == RolloutStrategyAttributeConditional.Equals:
            return next(filter(lambda x: ip_address(supplied_value) in ip_network(x), attr.values), None) is not None
        elif attr.conditional == RolloutStrategyAttributeConditional.Excludes \
                or attr.conditional == RolloutStrategyAttributeConditional.NotEquals:
            return next(filter(lambda x: ip_address(supplied_value) in ip_network(x), attr.values), None) is None

        return False


class MatcherRegistry(MatcherRepository):
    def find_matcher(self, attr: RolloutStrategyAttribute) -> StrategyMatcher:
        if attr.field_type == RolloutStrategyFieldType.String or \
                attr.field_type == RolloutStrategyFieldType.Date or \
                attr.field_type == RolloutStrategyFieldType.Datetime:
            return StringMatcher()

        if attr.field_type == RolloutStrategyFieldType.SemanticVersion:
            return SemanticVersionMatcher()

        if attr.field_type == RolloutStrategyFieldType.Number:
            return NumberMatcher()

        if attr.field_type == RolloutStrategyFieldType.Boolean:
            return BooleanMatcher()

        if attr.field_type == RolloutStrategyFieldType.IpAddress:
            return IPNetworkMatcher()

        return FallthroughMatcher()


class ApplyFeature:
    _percentageCalculator: PercentageCalculator
    _matcherRepository: MatcherRepository

    def __init__(self, percentage_calculator: Optional[PercentageCalculator] = None,
                 matcher_repository: Optional[MatcherRepository] = None):
        self._percentageCalculator = percentage_calculator if percentage_calculator is not None \
            else Murmur3PercentageCalculator()
        self._matcherRepository = matcher_repository if matcher_repository is not None else MatcherRegistry()

    def apply(self, strategies: list[RolloutStrategy], key: str, feature_value_id: str,
              context: ClientContext) -> Applied:
        if context is None or strategies is None or len(strategies) == 0:
            return Applied(False, None)

        percentage: Optional[float] = None
        percentage_key: Optional[str] = None
        base_percentage: dict[str, float] = {}
        default_percentage_key = context.default_percentage_key

        for rsi in strategies:
            if rsi.percentage != 0 and (default_percentage_key is not None or
                                        rsi.has_percentage_attributes):
                new_percentage_key = self.determine_percentage_key(context, rsi)

                if new_percentage_key not in base_percentage:
                    base_percentage[new_percentage_key] = 0

                base_percentage_val = base_percentage.get(new_percentage_key)

                if percentage is None or new_percentage_key != percentage_key:
                    percentage_key = new_percentage_key
                    percentage = self._percentageCalculator.determine_client_percentage(percentage_key, feature_value_id)

                    use_base_percentage = 0 if rsi.has_attributes else base_percentage_val

                    # if the percentage is lower than the user's key/feature-id then apply it
                    if percentage <= (use_base_percentage + rsi.percentage):
                        if rsi.has_attributes:
                            if self.match_attribute(context, rsi):
                                return Applied(True, rsi.value)
                        else:
                            return Applied(True, rsi.value)

                    if not rsi.has_attributes:
                        base_percentage[percentage_key] = base_percentage.get(percentage_key) + rsi.percentage

            if rsi.percentage == 0 and rsi.has_attributes and self.match_attribute(context, rsi):
                print('matched here')
                return Applied(True, rsi.value)

        print('here')
        return Applied(False, None)

    def match_attribute(self, context: ClientContext, rs: RolloutStrategy) -> bool:
        for attr in rs.attributes:
            supplied_value = context.get_attr(attr.field_name, None)
            if supplied_value is None and attr.field_name.lower() == 'now':
                if attr.field_type == RolloutStrategyFieldType.Date:
                    supplied_value = datetime.datetime.utcnow().isoformat()[0:10]
                elif attr.field_type == RolloutStrategyFieldType.Datetime:
                    supplied_value = datetime.datetime.utcnow().isoformat()

            if attr.values is None or supplied_value is None:
                if attr.conditional != RolloutStrategyAttributeConditional.Equals:
                    return False

                continue  # skip this loop

            if attr.values is None or supplied_value is None:
                return False

            if not self._matcherRepository.find_matcher(attr).match(supplied_value, attr):
                return False

        return True

    @staticmethod
    def determine_percentage_key(context: ClientContext, rs: RolloutStrategy) -> str:
        if not rs.has_percentage_attributes:
            return context.default_percentage_key

        return "$".join(map(lambda x:
                            context.get_attr(x, '<none>')[0] if isinstance(context.get_attr(x, '<none>'), list)
                            else context.get_attr(x, '<none>'), rs.percentage_attributes))




from enum import Enum


class StrategyAttributeDeviceName(Enum):
    Browser = 'browser'
    Mobile = 'mobile'
    Desktop = 'desktop'
    Server = 'server'
    Watch = 'watch'
    Embedded = 'embedded'



class EdgeService:
    async def poll(self):
        pass

    def client_evaluated(self):
        pass

    def close(self):
        pass

    async def context_change(self, header: str):
        pass



from __future__ import annotations

import typing
import os
import logging
import sys
from featurehub_sdk.client_context import ClientContext, ClientEvalFeatureContext, ServerEvalFeatureContext, \
    InternalFeatureRepository
from featurehub_sdk.edge_service import EdgeService
from featurehub_sdk.featurehub_repository import FeatureHubRepository
from collections.abc import Callable

from featurehub_sdk.polling_edge_service import PollingEdgeService
from featurehub_sdk.streaming_edge_service import StreamingEdgeClient

log = logging.getLogger(sys.modules[__name__].__name__)

class FeatureHubConfig:
    _api_keys: list[str]
    _edge_url: str
    _client_eval: bool
    _repository: InternalFeatureRepository
    _edge_service: typing.Optional[EdgeService]
    _edge_service_provider: Callable[[InternalFeatureRepository, list[str], str], EdgeService]

    def __init__(self, edge_url, api_keys: list[str],
                 repository: typing.Optional[InternalFeatureRepository] = None,
                 edge_provider: typing.Optional[Callable[[InternalFeatureRepository, list[str], str], EdgeService]] = None):
        self._edge_service = None
        self._repository = repository if repository is not None else FeatureHubRepository()
        self._edge_url = edge_url
        self._api_keys = api_keys

        if not edge_url or not api_keys:
            raise TypeError('api_key and edge_url must not be null')

        if any("*" in key for key in api_keys):
            if not all("*" in key for key in api_keys):
                raise TypeError('all api keys provided must be of the same type - all keys client or all keys eval')

        self._client_eval = '*' in self._api_keys[0]

        if not self._edge_url[-1] == '/':
            self._edge_url += '/'

        # this is the function we use to create our edge service if no other is specified
        self._edge_service_provider = edge_provider if edge_provider else self._create_default_provider
        self._edge_service = None

    def client_evaluated(self) -> bool:
        return self._client_eval  # is this correct?

    def get_api_keys(self) -> list[str]:
        return self._api_keys.copy()

    def get_host(self) -> str:
        return self._edge_url

    def repository(self, repository: typing.Optional[FeatureHubRepository] = None) -> InternalFeatureRepository:
        if repository:
            self._repository = repository

        return self._repository

    async def init(self) -> FeatureHubConfig:
        log.info("Init request made")
        # ensure the repository exists
        self.repository()

        # ensure the edge service provider exists
        await self.get_or_create_edge_service().poll()

        return self

    # uses the defined provider to make us an edge instance
    def _create_edge_service(self) -> EdgeService:
        # call to get the edge service method and then call that method with the parameters
        # we don't store this result as Server Evaluated clients need new
        return self.edge_service_provider()(self._repository, self._api_keys, self._edge_url)

    # checks to see if we already have an edge service, and if we don't, ask the provider to create
    # one for us
    def get_or_create_edge_service(self) -> EdgeService:
        if self._edge_service is None:
            self._edge_service = self._create_edge_service()
        return self._edge_service

    # this returns a function pointer that we call to get the actual edge service instance
    # it allows people to replace the way we create the edge service with a method of their own, e.g. with
    # a different provider or a customised provider creation
    def edge_service_provider(self,
                              edge_provider:
                              typing.Optional[Callable[[InternalFeatureRepository, list[str], str], EdgeService]] = None) \
            -> Callable[[InternalFeatureRepository, list[str], str], EdgeService]:
        # did the give us a new one? if not, return the default one we created on init
        if edge_provider is None:
            return self._edge_service_provider

        self._edge_service_provider = edge_provider

        # they have changed the provider, so make sure its closed and
        if self._edge_service:
            self._edge_service.close()
            self._edge_service = None

        return self._edge_service_provider

    # this is just an internal function that creates the default edge service if the user hasn't provided one (which
    # will be 90% of the time)
    def _create_default_provider(self, repository: InternalFeatureRepository, api_keys: list[str],
                                 edge_url: str) -> EdgeService:

        return StreamingEdgeClient(edge_url, api_keys, repository)

    def use_polling_edge_service(self, interval: int = int(os.environ.get("FEATUREHUB_POLL_INTERVAL", "30"))):
        self.edge_service_provider(lambda repository, api_keys,
                                          edge_url: PollingEdgeService(edge_url, api_keys, repository, interval))

    def new_context(self) -> ClientContext:
        repository = self.repository()
        edge_service = self.get_or_create_edge_service()

        return ClientEvalFeatureContext(repository, edge_service) \
            if self._client_eval else \
            ServerEvalFeatureContext(repository, edge_service)

    def close(self):
        if self._edge_service is not None:
            self._edge_service.close()
            self._edge_service = None



from typing import Optional

import sseclient
import urllib3
import json
import threading
import logging
import sys

from featurehub_sdk.client_context import InternalFeatureRepository
from featurehub_sdk.edge_service import EdgeService
from featurehub_sdk.version import sdk_version

log = logging.getLogger(sys.modules[__name__].__name__)


class _StreamingThread(threading.Thread):
    _cancel: bool
    _http: urllib3.PoolManager
    _client: Optional[sseclient.SSEClient] = None
    _repository: InternalFeatureRepository
    _url: str

    def __init__(self, edge_url: str, api_keys: list[str],
                 repository: InternalFeatureRepository):
        super().__init__(daemon=True, name="streaming-featurehub")
        self._url = f"{edge_url}/features/{api_keys[0]}"
        self._repository = repository
        self._cancel = False
        self._http = urllib3.PoolManager()

    def run(self):
        headers = {'Accept': 'text/event-stream', 'X-SDK': 'Python', 'X-SDK-Version': sdk_version}
        # headers = {'Accept': 'text/event-stream'}
        last_event_id = None
        while not self._cancel:
            try:
                log.log(5, "featurehub starting request: %s", self._url)
                if last_event_id is not None:
                    headers['Last-Event-Id'] = last_event_id
                resp = self._http.request('GET', self._url, preload_content=False, headers=headers)
                if resp.status == 200:
                    self._client = sseclient.SSEClient(resp)
                    for event in self._client.events():
                        last_event_id = event.id
                        print("data is " + event.event, event.data)
                        log.log(5, "received data %s: %s", event.event, event.data)
                        self._repository.notify(event.event, self._check_data(event.data))
                        if self._cancel:
                            self._client.close()
                elif resp.status == 404:
                    self._cancel = True
            except (ValueError, urllib3.exceptions.ProtocolError):
                pass

    def _check_data(self, data):
        if data and (data.startswith('{') or data.startswith('[')):
            return json.loads(data)

        return data

    def cancel(self):
        self._cancel = True

    def close(self):
        self._cancel = True
        if self._client:
            self._client.close()


class StreamingEdgeClient(EdgeService):
    _client_evaluated: bool
    _streaming_thread: _StreamingThread

    def __init__(self, edge_url: str, api_keys: list[str],
                 repository: InternalFeatureRepository):
        self._streaming_thread = _StreamingThread(edge_url, api_keys, repository)

        self._client_evaluated = '*' in api_keys[0]

    async def poll(self):
        if self._streaming_thread.is_alive():
            return

        self._streaming_thread.start()

    def close(self):
        if self._streaming_thread.is_alive():
            self._streaming_thread.close()

    def client_evaluated(self):
        return self._client_evaluated

    # not supported
    async def context_change(self, header: str):
        pass


from decimal import Decimal
from typing import Optional, Union

from featurehub_sdk.client_context import ClientContext, FeatureState, InternalFeatureRepository, RolloutStrategy


# this represents one of 3 things:
# - a legtimate feature state set of data as it has come from the server
# - an empty holder because the user has asked for a feature state that does not exist and we need to
#   provide them with an API that doesn't explode
# - a single hierarchy tree of a feature. When a context is created it needs to hold onto the individual context
#   and teh feature state it refers to, as many different concurrent users could have an active context pointing to
#   that feature. so ctx1 and ctx2 will each get their own copy of a FeatureStateHolder but it stores no actual state,
#   only a reference to the parent holder, that means as new data comes in, the context needs to always refer to the
#   current state, we have no way of searching for all of the instances of context/features and updating them.
from featurehub_sdk.interceptors import InterceptorValue


class FeatureStateHolder(FeatureState):
    """Holder for features. Wraps raw response with features dictionary"""

    _key: Optional[str]
    _internal_feature_state = None
    _parent_state: "FeatureStateHolder"
    _ctx: ClientContext
    _repo: InternalFeatureRepository
    _encoded_strategies: list[RolloutStrategy]

    # we can be initialised with no state when someone request a key that does not exist
    # the parent exists so we can keep track of the original feature when we use contexts
    def __init__(self, key: str,
                 repo: InternalFeatureRepository,
                 feature_state: Optional = None,
                 parent_state: Optional["FeatureStateHolder"] = None,
                 ctx: Optional[ClientContext] = None,
                 ):
        super().__init__()

        self._key = key
        self._parent_state = parent_state
        self._ctx = ctx
        self._repo = repo
        self._encoded_strategies = []

        if feature_state:
            self.__set_feature_state(feature_state)

    def __set_feature_state(self, feature_state):
        self._internal_feature_state = feature_state
        found_strategies = feature_state.get('strategies') if feature_state and feature_state.get('strategies') else []

        self._encoded_strategies = list(map(lambda rs: RolloutStrategy(rs), found_strategies))

    def __get_value(self, feature_type: Optional[str]) -> Union[None, bool, str, float]:
        if not self.locked:
            intercept = self._repo.find_interceptor(self._key)

            if intercept:
                return intercept.cast(feature_type if feature_type else 'STRING')

        # walk up the chain to find the original feature state (if any)
        fs = self._top_feature_state_holder()

        state = fs._feature_state()

        if state is None:
            return None

        # if the feature isn't a feature (they have asked for a feature that doesn't exist
        # or the type is wrong, return None
        if fs is None or (feature_type is not None and fs.feature_type != feature_type):
            return None

        if self._ctx is not None:
            matched = self._repo.apply(fs._encoded_strategies, self._key, fs.id, self._ctx)

            if matched.matched:
                return InterceptorValue(matched.value).cast(feature_type)

        return fs._feature_state().get('value')

    def with_context(self, ctx: ClientContext) -> FeatureState:
        return FeatureStateHolder(self._key, self._repo, None, self, ctx)

    def _top_feature_state_holder(self) -> "FeatureStateHolder":
        if self._parent_state:
            return self._parent_state._top_feature_state_holder()

        return self

    def _feature_state(self) -> dict:
        return self._top_feature_state_holder()._internal_feature_state

    @property
    def id(self) -> Optional[str]:
        return self._feature_state().get('id') if self.exists else None

    @property
    def feature_type(self) -> Optional[str]:
        return self._feature_state().get('type') if self.exists else None

    @property
    def locked(self):
        fs = self._feature_state()
        return fs.get('l') if fs else False

    def set_feature_state(self, feature_state: Optional[dict]):
        self.__set_feature_state(feature_state)

    @property
    def get_value(self):
        return self.__get_value(None)

    @property
    def get_version(self) -> int:
        fs = self._feature_state()
        return fs.get('version') if fs else -1

    @property
    def key(self) -> str:
        return self._key

    @property
    def get_string(self) -> Optional[str]:
        return self.__get_value('STRING')

    @property
    def get_number(self) -> Optional[Decimal]:
        return self.__get_value('NUMBER')

    @property
    def get_raw_json(self) -> Optional[str]:
        return self.__get_value('JSON')

    @property
    def get_boolean(self) -> Optional[bool]:
        return self.__get_value('BOOLEAN')

    @property
    def exists(self) -> bool:
        f = self._feature_state()
        return (f.get('l') is not None) if f else False

    @property
    def get_flag(self) -> Optional[bool]:
        return self.get_boolean

    @property
    def is_enabled(self) -> bool:
        return self.get_boolean is True

    @property
    def is_set(self) -> bool:
        return self.__get_value(self.feature_type) is not None

    def _get_internal_feature_state(self):
        return self._internal_feature_state if self.exists else None

    internal_feature_state = property(_get_internal_feature_state, __set_feature_state)


from enum import Enum


class StrategyAttributeCountryName(Enum):
    Afghanistan = 'afghanistan'
    Albania = 'albania'
    Algeria = 'algeria'
    Andorra = 'andorra'
    Angola = 'angola'
    AntiguaAndBarbuda = 'antigua_and_barbuda'
    Argentina = 'argentina'
    Armenia = 'armenia'
    Australia = 'australia'
    Austria = 'austria'
    Azerbaijan = 'azerbaijan'
    TheBahamas = 'the_bahamas'
    Bahrain = 'bahrain'
    Bangladesh = 'bangladesh'
    Barbados = 'barbados'
    Belarus = 'belarus'
    Belgium = 'belgium'
    Belize = 'belize'
    Benin = 'benin'
    Bhutan = 'bhutan'
    Bolivia = 'bolivia'
    BosniaAndHerzegovina = 'bosnia_and_herzegovina'
    Botswana = 'botswana'
    Brazil = 'brazil'
    Brunei = 'brunei'
    Bulgaria = 'bulgaria'
    BurkinaFaso = 'burkina_faso'
    Burundi = 'burundi'
    CaboVerde = 'cabo_verde'
    Cambodia = 'cambodia'
    Cameroon = 'cameroon'
    Canada = 'canada'
    CentralAfricanRepublic = 'central_african_republic'
    Chad = 'chad'
    Chile = 'chile'
    China = 'china'
    Colombia = 'colombia'
    Comoros = 'comoros'
    CongoDemocraticRepublicOfThe = 'congo_democratic_republic_of_the'
    CongoRepublicOfThe = 'congo_republic_of_the'
    CostaRica = 'costa_rica'
    CoteDivoire = 'cote_divoire'
    Croatia = 'croatia'
    Cuba = 'cuba'
    Cyprus = 'cyprus'
    CzechRepublic = 'czech_republic'
    Denmark = 'denmark'
    Djibouti = 'djibouti'
    Dominica = 'dominica'
    DominicanRepublic = 'dominican_republic'
    EastTimor = 'east_timor'
    Ecuador = 'ecuador'
    Egypt = 'egypt'
    ElSalvador = 'el_salvador'
    EquatorialGuinea = 'equatorial_guinea'
    Eritrea = 'eritrea'
    Estonia = 'estonia'
    Eswatini = 'eswatini'
    Ethiopia = 'ethiopia'
    Fiji = 'fiji'
    Finland = 'finland'
    France = 'france'
    Gabon = 'gabon'
    TheGambia = 'the_gambia'
    Georgia = 'georgia'
    Germany = 'germany'
    Ghana = 'ghana'
    Greece = 'greece'
    Grenada = 'grenada'
    Guatemala = 'guatemala'
    Guinea = 'guinea'
    GuineaBissau = 'guinea_bissau'
    Guyana = 'guyana'
    Haiti = 'haiti'
    Honduras = 'honduras'
    Hungary = 'hungary'
    Iceland = 'iceland'
    India = 'india'
    Indonesia = 'indonesia'
    Iran = 'iran'
    Iraq = 'iraq'
    Ireland = 'ireland'
    Israel = 'israel'
    Italy = 'italy'
    Jamaica = 'jamaica'
    Japan = 'japan'
    Jordan = 'jordan'
    Kazakhstan = 'kazakhstan'
    Kenya = 'kenya'
    Kiribati = 'kiribati'
    KoreaNorth = 'korea_north'
    KoreaSouth = 'korea_south'
    Kosovo = 'kosovo'
    Kuwait = 'kuwait'
    Kyrgyzstan = 'kyrgyzstan'
    Laos = 'laos'
    Latvia = 'latvia'
    Lebanon = 'lebanon'
    Lesotho = 'lesotho'
    Liberia = 'liberia'
    Libya = 'libya'
    Liechtenstein = 'liechtenstein'
    Lithuania = 'lithuania'
    Luxembourg = 'luxembourg'
    Madagascar = 'madagascar'
    Malawi = 'malawi'
    Malaysia = 'malaysia'
    Maldives = 'maldives'
    Mali = 'mali'
    Malta = 'malta'
    MarshallIslands = 'marshall_islands'
    Mauritania = 'mauritania'
    Mauritius = 'mauritius'
    Mexico = 'mexico'
    MicronesiaFederatedStatesOf = 'micronesia_federated_states_of'
    Moldova = 'moldova'
    Monaco = 'monaco'
    Mongolia = 'mongolia'
    Montenegro = 'montenegro'
    Morocco = 'morocco'
    Mozambique = 'mozambique'
    Myanmar = 'myanmar'
    Namibia = 'namibia'
    Nauru = 'nauru'
    Nepal = 'nepal'
    Netherlands = 'netherlands'
    NewZealand = 'new_zealand'
    Nicaragua = 'nicaragua'
    Niger = 'niger'
    Nigeria = 'nigeria'
    NorthMacedonia = 'north_macedonia'
    Norway = 'norway'
    Oman = 'oman'
    Pakistan = 'pakistan'
    Palau = 'palau'
    Panama = 'panama'
    PapuaNewGuinea = 'papua_new_guinea'
    Paraguay = 'paraguay'
    Peru = 'peru'
    Philippines = 'philippines'
    Poland = 'poland'
    Portugal = 'portugal'
    Qatar = 'qatar'
    Romania = 'romania'
    Russia = 'russia'
    Rwanda = 'rwanda'
    SaintKittsAndNevis = 'saint_kitts_and_nevis'
    SaintLucia = 'saint_lucia'
    SaintVincentAndTheGrenadines = 'saint_vincent_and_the_grenadines'
    Samoa = 'samoa'
    SanMarino = 'san_marino'
    SaoTomeAndPrincipe = 'sao_tome_and_principe'
    SaudiArabia = 'saudi_arabia'
    Senegal = 'senegal'
    Serbia = 'serbia'
    Seychelles = 'seychelles'
    SierraLeone = 'sierra_leone'
    Singapore = 'singapore'
    Slovakia = 'slovakia'
    Slovenia = 'slovenia'
    SolomonIslands = 'solomon_islands'
    Somalia = 'somalia'
    SouthAfrica = 'south_africa'
    Spain = 'spain'
    SriLanka = 'sri_lanka'
    Sudan = 'sudan'
    SudanSouth = 'sudan_south'
    Suriname = 'suriname'
    Sweden = 'sweden'
    Switzerland = 'switzerland'
    Syria = 'syria'
    Taiwan = 'taiwan'
    Tajikistan = 'tajikistan'
    Tanzania = 'tanzania'
    Thailand = 'thailand'
    Togo = 'togo'
    Tonga = 'tonga'
    TrinidadAndTobago = 'trinidad_and_tobago'
    Tunisia = 'tunisia'
    Turkey = 'turkey'
    Turkmenistan = 'turkmenistan'
    Tuvalu = 'tuvalu'
    Uganda = 'uganda'
    Ukraine = 'ukraine'
    UnitedArabEmirates = 'united_arab_emirates'
    UnitedKingdom = 'united_kingdom'
    UnitedStates = 'united_states'
    Uruguay = 'uruguay'
    Uzbekistan = 'uzbekistan'
    Vanuatu = 'vanuatu'
    VaticanCity = 'vatican_city'
    Venezuela = 'venezuela'
    Vietnam = 'vietnam'
    Yemen = 'yemen'
    Zambia = 'zambia'
    Zimbabwe = 'zimbabwe'



from setuptools import setup

from featurehub_sdk.version import sdk_version

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='featurehub-sdk',
    version=sdk_version,
    license='MIT',
    description='Official Python SDK for FeatureHub',
    url='https://github.com/featurehub-io/featurehub-python-sdk',
    author='FeatureHub.io',
    author_email='info@featurehub.io',
    keywords='feature-flags, feature-toggles, flags, toggles',
    packages=['featurehub_sdk'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=['urllib3==1.26.*',
                      'sseclient-py==1.7.*',
                      'murmurhash2==0.2.*',
                      'node_semver==0.8.*'],
)



from __future__ import annotations

from decimal import Decimal
from enum import Enum
from typing import Optional, Any
import json
import urllib.parse
import asyncio

from featurehub_sdk.edge_service import EdgeService
from featurehub_sdk.interceptors import InterceptorValue
from featurehub_sdk.strategy_attribute_country_name import StrategyAttributeCountryName
from featurehub_sdk.strategy_attribute_device_name import StrategyAttributeDeviceName
from featurehub_sdk.strategy_attribute_platform_name import StrategyAttributePlatformName


class FeatureState:
    # Python can't cope with circular dependencies, which is why this is here, FeatureState/Context are a tree
    # structure of each other, e.g. feature('X') -> raw value, ctx.feature('X').withContext(ctx).get_boolean()
    @property
    def get_value(self):
        return ""

    @property
    def locked(self):
        return False

    @property
    def id(self):
        return ""

    @property
    def get_version(self) -> int:
        return -1

    @property
    def get_key(self) -> str:
        return ""

    @property
    def get_string(self) -> Optional[str]:
        return None

    @property
    def get_number(self) -> Optional[Decimal]:
        return None

    @property
    def get_raw_json(self) -> Optional[str]:
        return None

    @property
    def get_boolean(self) -> Optional[bool]:
        return None

    @property
    def get_flag(self) -> Optional[bool]:
        return None

    @property
    def is_enabled(self) -> bool:
        return False

    @property
    def is_set(self) -> bool:
        return False

    @property
    def exists(self) -> bool:
        return False

    @property
    def key(self) -> str:
        return ""

    def with_context(self, ctx: ClientContext) -> "FeatureState":
        pass

# the RolloutX are normally generated from OpenAPI


class RolloutStrategyAttributeConditional(Enum):
    Equals = 'EQUALS'
    EndsWith = 'ENDS_WITH'
    StartsWith = 'STARTS_WITH'
    Greater = 'GREATER'
    GreaterEquals = 'GREATER_EQUALS'
    Less = 'LESS'
    LessEquals = 'LESS_EQUALS'
    NotEquals = 'NOT_EQUALS'
    Includes = 'INCLUDES'
    Excludes = 'EXCLUDES'
    Regex = 'REGEX'


class RolloutStrategyFieldType(Enum):
    String = 'STRING'
    SemanticVersion = 'SEMANTIC_VERSION'
    Number = 'NUMBER'
    Date = 'DATE'
    Datetime = 'DATETIME'
    Boolean = 'BOOLEAN'
    IpAddress = 'IP_ADDRESS'


class RolloutStrategyAttribute:
    _attr: dict

    def __init__(self, attr: dict):
        self._attr = attr

    @property
    def id(self) -> Optional[str]:
        return self._attr.get('id')

    @property
    def conditional(self) -> RolloutStrategyAttributeConditional:
        return RolloutStrategyAttributeConditional(self._attr.get('conditional'))

    @property
    def field_name(self) -> str:
        return self._attr.get('fieldName')

    @property
    def values(self) -> Optional[list[Any]]:
        return self._attr.get('values')

    @property
    def float_values(self) -> list[float]:
        return list(map(lambda x: float(x), filter(lambda x: x is not None, self.values)))

    @property
    def str_values(self) -> list[str]:
        return list(map(lambda x: str(x), filter(lambda x: x is not None, self.values)))

    @property
    def field_type(self) -> RolloutStrategyFieldType:
        return RolloutStrategyFieldType(self._attr.get('type'))


class RolloutStrategy:
    _attr: dict
    _attributes: list[RolloutStrategyAttribute]

    def __init__(self, attr: dict):
        self._attr = attr

        rsa = self._attr.get('attributes')
        self._attributes = list(map(lambda x: RolloutStrategyAttribute(x), list(rsa))) if rsa else []

    @property
    def id(self) -> Optional[str]:
        return self._attr.get('id')

    @property
    def name(self) -> Optional[str]:
        return self._attr.get('name')

    @property
    def percentage(self) -> int:
        p = self._attr.get('percentage')
        return int(p) if p is not None else 0

    @property
    def percentage_attributes(self) -> list[str]:
        pa = self._attr.get('percentageAttributes')

        return list(pa) if pa is not None else []

    @property
    def has_percentage_attributes(self) -> bool:
        pa = self._attr.get('percentageAttributes')
        return len(pa) > 0 if pa is not None else False

    @property
    def value(self) -> Optional[Any]:
        return self._attr.get('value')

    @property
    def attributes(self) -> list[RolloutStrategyAttribute]:
        return self._attributes

    @property
    def has_attributes(self) -> bool:
        return len(self._attributes) > 0

class Applied:
    _matched: bool
    _value: Any

    def __init__(self, matched: bool, value: Any):
        self._matched = matched
        self._value = value

    @property
    def value(self):
        return self._value

    @property
    def matched(self):
        return self._matched


class InternalFeatureRepository:
    # give me the raw feature from the repository level (top level features, no contexts)
    def feature(self, key: str) -> FeatureState:
        pass

    # is there an interceptor for this feature value?
    def find_interceptor(self, feature_value: str) -> Optional[InterceptorValue]:
        pass

    # is the repo ready? has it had at least one set of features from any source?
    def is_ready(self) -> bool:
        pass

    # set repo to not being ready
    def not_ready(self) -> bool:
        pass

    def apply(self, strategies: list[RolloutStrategy], key: str, feature_id: str, context: "ClientContext") -> Applied:
        pass

    def notify(self, cmd: str, data):
        pass

class ClientContext:
    """holds client context"""
    _attributes: dict[str, object]
    _repo: InternalFeatureRepository
    USER_KEY = 'userkey'
    SESSION = 'session'
    COUNTRY = 'country'
    DEVICE = 'device'
    PLATFORM = 'platform'
    VERSION = 'version'

    def __init__(self, repo: InternalFeatureRepository):
        self._repository = repo
        self._attributes = {}

    def user_key(self, value: str) -> ClientContext:
        self._attributes[ClientContext.USER_KEY] = value
        return self

    def session_key(self, value: str) -> ClientContext:
        self._attributes[ClientContext.SESSION] = value
        return self

    def country(self, value: StrategyAttributeCountryName) -> ClientContext:
        self._attributes[ClientContext.COUNTRY] = value
        return self

    def device(self, value: StrategyAttributeDeviceName) -> ClientContext:
        self._attributes[ClientContext.DEVICE] = value
        return self

    def platform(self, value: StrategyAttributePlatformName) -> ClientContext:
        self._attributes[ClientContext.PLATFORM] = value
        return self

    def version(self, version: str) -> ClientContext:
        self._attributes[ClientContext.VERSION] = version
        return self

    def attribute_values(self, key: str, values: list[str]) -> ClientContext:
        self._attributes[key] = values
        return self

    def clear(self) -> ClientContext:
        self._attributes.clear()
        return self

    def get_attr(self, key: str, default_value: Optional[str] = None) -> Optional[object]:
        if key in self._attributes:
            return self._attributes.get(key)
        return default_value

    @property
    def default_percentage_key(self) -> str:
        return self.get_attr(ClientContext.SESSION) if self.get_attr(ClientContext.SESSION) \
            else str(self.get_attr(ClientContext.USER_KEY))

    def is_enabled(self, name: str) -> bool:
        return self.feature(name).is_enabled

    def feature(self, name: str) -> FeatureState:
        # context never matters as the repository always reflects the correctly evaluated state
        return self._repository.feature(name)

    def is_set(self, name: str) -> bool:
        return self.feature(name).is_set

    def get_number(self, name: str) -> Optional[Decimal]:
        return self.feature(name).get_number

    def get_string(self, name: str) -> Optional[str]:
        return self.feature(name).get_string

    def get_json(self, name: str) -> Optional[any]:
        val = self.feature(name).get_raw_json
        return json.loads(val) if val else None

    def get_raw_json(self, name: str) -> Optional[str]:
        return self.feature(name).get_raw_json

    def get_flag(self, name: str) -> Optional[bool]:
        return self.feature(name).get_flag

    def get_boolean(self, name: str) -> Optional[bool]:
        return self.feature(name).get_boolean

    async def build(self) -> ClientContext:
        pass

    def build_sync(self) -> ClientContext:
        pass

    async def close(self):
        pass


class ClientEvalFeatureContext(ClientContext):
    _edge: EdgeService

    def __init__(self, repo: InternalFeatureRepository, edge: EdgeService):
        super().__init__(repo)
        self._edge = edge

    async def build(self) -> ClientContext:
        await self._edge.poll()
        return self

    def build_sync(self) -> ClientContext:
        # assumes you have already dont an init yourself and the repository is up and going, you should
        # use this on your server instances
        return self

    async def close(self):
        self._edge.close()

    def feature(self, name: str) -> FeatureState:
        return self._repository.feature(name).with_context(self)


class ServerEvalFeatureContext(ClientContext):
    # server eval feature context needs to evaluate the context on the server, so we need to wrap up the
    # context in a bunch of url encoded key value pairs and send it off to the edge service to be updated and
    # refreshed
    _old_header: Optional[str] = None
    _current_edge: Optional[EdgeService]

    def __init__(self, repo: InternalFeatureRepository, edge: EdgeService):
        super().__init__(repo)
        self._current_edge = edge

    def _pick_first(self, v: object):
        return v[0] if isinstance(v, list) else v

    async def build(self) -> ClientContext:
        new_header = "&".join("=".join((k, urllib.parse.quote(str(self._pick_first(v))))) for k,v in self._attributes.items())

        if self._old_header is None and len(new_header) == 0:
            await self._current_edge.poll()  # just make sure we have started
        elif new_header != self._old_header: # make sure it changed
            self._old_header = new_header

            self._repository.not_ready()

            await self._current_edge.context_change(new_header)

        return self

    def build_sync(self) -> ClientContext:
        asyncio.run(self.build())
        return self

    async def close(self):
        if self._current_edge:
            self._current_edge.close()
            self._current_edge = None
            self._old_header = None



from typing import Optional

from featurehub_sdk.client_context import InternalFeatureRepository, ClientContext, Applied, RolloutStrategy
from featurehub_sdk.fh_state_base_holder import FeatureStateHolder
from featurehub_sdk.interceptors import ValueInterceptor, InterceptorValue
from featurehub_sdk.strategy_matchers import ApplyFeature


class FeatureHubRepository(InternalFeatureRepository):
    features: dict[str, FeatureStateHolder] # do we need this to be private and expose it as a getter method?
    _ready: bool = False
    _interceptors: list[ValueInterceptor]
    _strategy_matcher: ApplyFeature

    def __init__(self, apply_features: Optional[ApplyFeature] = None):
        self._strategy_matcher = apply_features if apply_features is not None else ApplyFeature()
        self._interceptors = []
        self.features = {}

    def apply(self, strategies: list[RolloutStrategy], key: str, feature_id: str, context: ClientContext) -> Applied:
        return self._strategy_matcher.apply(strategies, key, feature_id, context)

    def notify(self, status: str, data: Optional):
        if status == 'failed':
            self._ready = False
            return

        if data is None:
            return

        if status == 'features':
            self.__update_features(data)
            self._ready = True
        elif status == 'feature':
            self.__update_feature_state(data)
            self._ready = True
        elif status == 'delete_feature':
            self._delete_feature(data)

    def _delete_feature(self, data: dict):
        feat = self.features.get(data['key'])
        if feat:
            feat.set_feature_state(None)

    def __update_features(self, data: list[dict]):
        if data:
            for feature_state in data:
                self.__update_feature_state(feature_state)

    def __update_feature_state(self, feature_state):
        if not feature_state or not feature_state.get('key'):
            return

        # check if feature already in the dictionary, if not add to the dictionary
        holder = self.features.get(feature_state['key'])
        if not holder:
            new_feature = FeatureStateHolder(feature_state['key'], self, feature_state, None, None)
            self.features[feature_state['key']] = new_feature
            return

        # if feature is in the dictionary, check if version has changed
        elif feature_state.get('version') < holder.get_version:
            return
        elif feature_state.get('version') == holder.get_version and feature_state.get('value') == holder.get_value:
            return

        holder.set_feature_state(feature_state)

    def is_ready(self):
        return self._ready

    def feature(self, key) -> FeatureStateHolder:
        # we follow the standard as per other SDKs, if the key doesn't exist, we create it so the
        # chain of repository.feature(X).get_flag() for example works but returns None and doesn't explode

        fs = self.features.get(key)

        if fs is None:
            fs = FeatureStateHolder(key, self, None, None, None)
            self.features[key] = fs

        return fs

    def not_ready(self):
        self._ready = False

    def register_interceptor(self, interceptor: ValueInterceptor):
        self._interceptors.append(interceptor)

    def find_interceptor(self, feature_value: str) -> Optional[InterceptorValue]:
        for interceptor in self._interceptors:
            found = interceptor.intercepted_value(feature_value)
            if found is not None:
                return found

    def extract_feature_state(self) -> list:
        # allows you to extract the internal state of value features out and store it outside the repository
        # if you wish
        feats = []

        for k, v in self.features.items():
            data = v.internal_feature_state
            if data:
                feats.append(data)

        return feats
